"""Interface for Service resources."""

import ipaddress

import pyrsistent

from kube import _base
from kube import _error
from kube import _meta
from kube import _util
from kube import _watch


class ServiceView(_base.ViewABC):
    """View of the Service resource items in the cluster.

    :param cluster: The cluster instance.
    :type cluster: kube.Cluster
    :param namespace: Limit the view to resource items in this
       namespace.
    :type namespace: str

    :ivar kube.Cluster cluster: The cluster instance this view is part of.
    :ivar str namespace: The namespace this view is contrained to, if
       the view is not bound to a namespace this will be ``None``.
    """

    def __init__(self, cluster, namespace=None):  # pylint: disable=super-init-not-called
        self.cluster = cluster
        self.namespace = namespace

    def __iter__(self):
        """Iterate over all Service resource items in the view.

        :returns: An object for each service in the view.
        :rtype: kube.ServiceItem

        :raises kube.APIError: For errors from the k8s API server.
        """
        if self.namespace:
            data = self.cluster.proxy.get('namespaces',
                                          self.namespace, 'services')
        else:
            data = self.cluster.proxy.get('services')
        for item in data['items']:
            yield ServiceItem(self.cluster, item)

    @property
    def kind(self):
        """The Kubernetes resource kind of the resource.

        This is a :class:`kube.Kind` enum.
        """
        return _base.Kind.ServiceList

    def fetch(self, name, namespace=None):
        """Retrieve an individual Service by name.

        This returns the current verison of the resource item.

        If the view itself is associated with a namespace, the
        :attr:`namespace` attribute is not ``None``, then this will
        default to fetching the Service resource item from this
        namespace.  Otherwise the ``namespace`` parameter must be
        supplied.

        :param name: The name of the Service to retrive.
        :type name: str
        :param namespace: Use a namespace different from
           :attr:`namespace`.
        :type namespace: str

        :returns: A :class:`kube.ServiceItem` instance.

        :raises LookupError: If the Service does not exist.
        :raises NamespaceError: When no namespace was supplied.
        :raises kube.APIError: For errors from the k8s API server.
        """
        ns = namespace or self.namespace
        if not ns:
            raise _error.NamespaceError('Missing namespace')
        data = _util.fetch_resource(
            self.cluster, 'services', name, namespace=ns)
        return ServiceItem(self.cluster, data)

    def filter(self, *, labels=None, fields=None):
        """Get a filtered iterator of Services.

        :param labels: A label selector expression.  This can either
           be a dictionary with labels which must match exactly, or a
           string label expression as understood by k8s itself.
        :type labels: dict or str
        :param fields: A field selector expression.  This can either
           be a dictionary with fields which must match exactly, or a
           string field selector as understood by k8s itself.
        :type fields: dict or str

        :returns: an iterator of :class:`SerivcesItem`s that match the
           given selector.

        :raises ValueError: If an empty selector is used.  An empty
           selector is almost certainly not what you want.  Kubernetes
           treats an **empty** selector as *all* items and treats a
           **null** selector as *no* items.
        :raises kube.APIError: for errors from the k8s API server.
        """
        data_iter = _util.filter_list(self.cluster, 'services',
                                      labels=labels, fields=fields,
                                      namespace=self.namespace)
        for item in data_iter:
            yield ServiceItem(self.cluster, item)

    def watch(self):
        """Watch the ServiceList of this view for changes to items.

        Whenever one of the resource items in the view changes a new
        :class:`kube.WatchEvent` instance is yielded.  You can
        currently not control from "when" resources are being watched,
        other then from "now".  So be aware of any race conditions
        with watching.

        :returns: An iterator of :class:`kube.WatchEvent` instances
            where the :attr:`kube.WatchEvent.object` attribute will be
            a :class:`ServiceItem` instance.

        :raises kube.NamespaceError: When the namespace no longer
           exists.
        :raises kube.APIError: For errors from the k8s API server.
        """
        if self.namespace:
            path = ['namespaces', self.namespace, 'services']
        else:
            path = ['services']
        svclist = self.cluster.proxy.get(*path)
        version = svclist['metadata']['resourceVersion']
        jsonwatcher = self.cluster.proxy.watch(*path, version=version)
        return _watch.ResourceWatcher(self.cluster, jsonwatcher, ServiceItem)


class ServiceItem(_base.ItemABC):
    """A Service in the Kubernetes cluster.

    :param cluster: The cluster this Service exists in.
    :type cluster: kube.Cluster
    :param raw: The raw data of the resource item.
    :type raw: pyrsistent.PMap

    :ivar cluster: The :class:`kube.Cluster` this Service exists in.
    :ivar raw: The raw decoded JSON object as returned by the API
       server.  This is a :class:`pyrsistent.PMap` instance.
    """

    def __init__(self, cluster, raw):  # pylint: disable=super-init-not-called
        self.cluster = cluster
        self.raw = raw

    @property
    def kind(self):
        """The Kubernetes resource kind of this resource.

        This is a :class:`kube.Kind` enum.
        """
        return _base.Kind.Service

    @property
    def meta(self):
        """The Service's metadata as a :class:`kube.ObjectMeta` instance."""
        return _meta.ObjectMeta(self)

    def spec(self):
        """The spec of this Service's resource item.

        This is a copy of the raw, decoded JSON data representing the
        spec of this Service.
        """
        return pyrsistent.thaw(self.raw['spec'])

    def watch(self):
        """Watch the Service for changes.

        Only changes after the current version will be part of the
        iterator.  However it can not be guaranteed that *every*
        change is returned, if the current version is rather old some
        changes might no longer be available.

        :returns: An iterator of :class:`kube.WatchEvent` instances
           where the :attr:`kube.WatchEvent.item` attribute will be
           a :class:`kube.ServiceItem` instance.

        :raises kube.APIError: For errors from the k8s API server.
        """
        return _util.watch_item(self)

    @property
    def loadbalancer_ingress(self):
        """The loadbalancer ingress endpoints.

        This is a set of ingress endpoints in use by the
        loadbalancer.  Depending on the infrastructure the cluster
        runs on the endpoint can be either an
        :class:`ipaddress.IPv4Address`, :class:`ipaddress.IPv6Address`
        or a hostname as a string.
        """
        try:
            ingress = self.raw['status']['loadBalancer']['ingress']
        except KeyError:
            raise _error.StatusError
        endpoints = set()
        for endpoint in ingress:
            if 'ip' in endpoint:
                endpoints.add(ipaddress.ip_address(endpoint['ip']))
            elif 'hostname' in endpoint:
                endpoints.add(endpoint['hostname'])
            else:
                raise _error.StatusError
        return endpoints



"""Pythonic Kubernetes API.

The main entrypoint to the API is to create a :class:`Cluster`
instance.  All other objects are normally created by an instance of
this :class:`Cluster` class.
"""

# pylint: disable=unused-import

from kube._error import APIError, StatusError, NamespaceError
from kube._base import ViewABC, ItemABC, Kind
from kube._meta import ObjectMeta, ResourceLabels
from kube._watch import WatchEventType, WatchEvent
from kube._cluster import Cluster, APIServerProxy
from kube._node import NodeView, NodeItem, AddressType
from kube._namespace import NamespaceView, NamespaceItem, NamespacePhase
from kube._pod import PodView, PodItem, PodPhase, Container, ContainerState
from kube._replicaset import ReplicaSetView, ReplicaSetItem
from kube._service import ServiceView, ServiceItem
from kube._secret import SecretView, SecretItem, SecretType



"""Tools to work with a Kubernetes cluster.

This module contains the toplevel tools to work with a Kubernets
cluster and it's API server.
"""

import urllib.parse

import requests

from kube import _error
from kube import _namespace
from kube import _node
from kube import _pod
from kube import _replicaset
from kube import _secret
from kube import _service
from kube import _util
from kube import _watch


class Cluster:
    """A Kubernetes cluster.

    The entrypoint to control a Kubernetes cluster.  There is only one
    connection mechanism, which is via a local API server proxy.  This
    is normally achieved by running ``kubectl proxy``.

    :param str url: The URL of the API server.

    :ivar proxy: A helper class to directly access the API server. An
        instance of :class:`kube.APIServerProxy` helper class.
    :ivar nodes: A :class:`kube.NodeView` instance.
    """

    def __init__(self, url='http://localhost:8001/api/'):
        if not url.endswith('/'):
            url += '/'
        api_url = urllib.parse.urljoin(url, 'v1/')
        self.proxy = APIServerProxy(api_url)
        self.namespaces = _namespace.NamespaceView(self)
        self.nodes = _node.NodeView(self)
        self.pods = _pod.PodView(self)
        self.replicasets = _replicaset.ReplicaSetView(self)
        self.services = _service.ServiceView(self)
        self.secrets = _secret.SecretView(self)

    def close(self):
        """Close and clean up underlying resources."""
        self.proxy.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_val, exc_type, traceback):
        self.close()


class APIServerProxy:
    """Helper class to directly communicate with the API server.

    Since most classes need to communicate with the Kubernetes
    cluster's API server in a common way this class helps take care of
    the common logic.  It also keeps the requests session alive to
    enable connection pooling to the API server.

    :param str base_url: The URL of the API, including the API version.
    """

    def __init__(self, base_url='http://localhost:8001/api/v1/'):
        if not base_url.endswith('/'):
            base_url += '/'
        self._base_url = base_url
        self._session = requests.Session()

    def close(self):
        """Close underlying connections.

        Once the proxy has been closed then the it can no longer be used
        to issue further requests.
        """
        self._session.close()

    def urljoin(self, *path):
        """Wrapper around urllib.parse.urljoin for the configured base URL.

        :param path: Individual relative path components, they will be
           joined using "/".  None of the path components should
           include a "/" separator themselves.
        """
        return urllib.parse.urljoin(self._base_url, '/'.join(path))

    def get(self, *path, **params):
        """HTTP GET the relative path from the API server.

        :param str path: Individual relative path components, they
           will be joined using "/".  None of the path components
           should include a "/" separator themselves.
        :param dict params: Extra query parameters for the URL of the
           GET request as a dictionary of strings.

        :returns: The decoded JSON data.
        :rtype: pyrsistent.PMap

        :raises kube.APIError: If the response status is not 200 OK.
        """
        url = self.urljoin(*path)
        response = self._session.get(url, params=params)
        if response.status_code != 200:
            raise _error.APIError(response, 'Failed to GET {}'.format(url))
        else:
            return response.json(cls=_util.ImmutableJSONDecoder)

    def post(self, *path, json=None, **params):
        """HTTP POST to the relative path on the API server.

        :param path: Individual relative path components, they will be
           joined using :meth:`urljoin`.
        :type path: str
        :param json: The body to post, which will be JSON-encoded
           before posting.
        :type json: collections.abc.Mapping
        :param params: Extra query parameters for the URL of the POST
           request.
        :type params: str

        :returns: The decoded JSON data.
        :rtype: pyrsistent.PMap

        :raises kube.APIError: If the response status is not 201
           Created.
        """
        url = self.urljoin(*path)
        response = self._session.post(url, json=json, params=params)
        if response.status_code != 201:
            raise _error.APIError(response, 'Failed to POST {}'.format(url))
        else:
            return response.json(cls=_util.ImmutableJSONDecoder)

    def delete(self, *path, json=None, **params):
        """HTTP DELETE to the relative path on the API server.

        :param path: Individual relative path components, they will be
           joined using :meth:`urljoin`.
        :type path: str
        :param json: The body, which will be JSON-encoded before
           posting.
        :type json: collections.abc.Mapping
        :param params: Extra query parameters for the URL of the
           DELETE request.
        :type params: str

        :returns: The decoded JSON data.
        :rtype: pyrsistent.PMap

        :raises kube.APIError: If the response status is not 200 OK.
        """
        url = self.urljoin(*path)
        response = self._session.delete(url, json=json, params=params)
        if response.status_code != 200:
            raise _error.APIError(response, 'Failed to DELETE {}'.format(url))
        else:
            return response.json(cls=_util.ImmutableJSONDecoder)

    def patch(self, *path, patch=None):
        """HTTP PATCH as application/strategic-merge-patch+json.

        This allows using the Strategic Merge Patch to patch a
        resource on the Kubernetes API server.

        :param str path: Individual relative path components, they
           will be joined using "/".  None of the path components
           should include a "/" separator themselves, unless you only
           provide one component which will be joined to the base URL
           using :func:`urllib.parse.urljoin`.  This case can be
           useful to use the links provided by the API itself
           directly, e.g. from a resource's ``metadata.selfLink``
           field.
        :param dict patch: The decoded JSON object with the patch
           data.

        :returns: The decoded JSON object of the resource after
           applying the patch.

        :raises APIError: If the response status is not 200 OK.
        """
        url = self.urljoin(*path)
        headers = {'Content-Type': 'application/strategic-merge-patch+json'}
        response = self._session.patch(url, headers=headers, json=patch)
        if response.status_code != 200:
            raise _error.APIError(response, 'Failed to PATCH {}:'.format(url))
        else:
            return response.json(cls=_util.ImmutableJSONDecoder)

    def watch(self, *path, version=None, fields=None):
        """Watch a list resource for events.

        This issues a request to the API with the ``watch`` query
        string parameter set to ``true`` which returns a chunked
        response.  An iterator is returned which continuously reads
        from the response, yielding received lines as bytes.

        :param path: The URL path to the resource to watch. See
            :meth:`urljoin`.
        :param str version: The resource version to start watching from.
        :param dict fields: A dict of fields which must match their
           values.  This is a limited form of the full fieldSelector
           format, it is limited because filtering is done at client
           side for consistecy.

        :returns: An special iterator which allows non-blocking
           iterating using a ``.next(timeout) method.  Using it as a
           normal iterator will result in blocking behaviour.
        :rtype: :class:`LineWatcher`.

        :raises APIError: If there is a problem with the API server.
        """
        return _watch.JSONWatcher(self, *path, version=version, fields=fields)



"""Interface for node resources."""

import collections
import enum
import ipaddress

import pyrsistent

from kube import _base
from kube import _error
from kube import _meta
from kube import _util
from kube import _watch


class NodeView(_base.ViewABC):
    """View of all the Node resource items in the cluster.

    :param cluster: The cluster instance.
    :type cluster: kube.Cluster
    :param namespace: Limit the view to resource items in this
       namespace.  This is here for the :class:`kube.ViewABC`
       compatibility but can not be used for the NodeList resource.  A
       :class:`kube.NamespaceError` is raised when this is not
       ``None``.
    :type namespace: NoneType

    :raises kube.NamespaceError: If instantiated using a namespace.

    :ivar cluster: The :class:`kube.Cluster` instance.
    """

    def __init__(self, cluster, namespace=None):  # pylint: disable=super-init-not-called
        if namespace is not None:
            raise _error.NamespaceError(
                'NodeList does not support namespaces')
        self.cluster = cluster
        self.namespace = None

    def __iter__(self):
        """Iterator over all Node resource items in the cluster.

        :raises kube.APIError: For errors from the k8s API server.
        """
        data = self.cluster.proxy.get('nodes')
        for item in data['items']:
            yield NodeItem(self.cluster, item)

    @property
    def kind(self):
        """The kind of the underlying Kubernetes list resource.

        This is a :class:`kube.Kind` enum.
        """
        return _base.Kind.NodeList

    def fetch(self, name, namespace=None):
        """Retrieve an individual node by name.

        This returns the current verison of the resource item.

        :param str name: The name of the node to retrieve.
        :param str namespace: Must be *None* or a
           :class:`kube.NamespaceError` is raised.  Here only for
           compatibility with the ABC.

        :return: A single :class:`kube.NodeItem` instance.
        :raises LookupError: If the node does not exist.
        :raises kube.APIError: For errors from the k8s API server.
        :raises kube.NamespaceError: If a namespace is used.
        """
        if namespace is not None:
            raise _error.NamespaceError(
                'NodeList does not support namespaces')
        data = _util.fetch_resource(self.cluster, 'nodes', name)
        return NodeItem(self.cluster, data)

    def filter(self, *, labels=None, fields=None):
        """Return an iterable of a subset of the resources.

        :param labels: A label selector expression.  This can either
           be a dictionary with labels which must match exactly, or a
           string label expression as understood by k8s itself.
        :type labels: dict or str
        :param fields: A field selector expression.  This can either
           be a dictionary with fields which must match exactly, or a
           string field selector as understood by k8s itself.
        :type fields: dict or str

        :returns: An iterator of :class:`kube.ItemABC` instances
           of the correct type for the resrouce.

        :raises ValueError: If an empty selector is used.  An empty
           selector is almost certainly not what you want.  Kubernetes
           treats an **empty** selector as *all* items and treats a
           **null** selector as *no* items.
        :raises kube.APIError: for errors from the k8s API server.
        """
        data_iter = _util.filter_list(self.cluster, 'nodes',
                                      labels=labels, fields=fields)
        for item in data_iter:
            yield NodeItem(self.cluster, item)

    def watch(self):
        """Watch the NodeList of this view for changes to items.

        Whenever one of the resources items in the view changes a new
        :class:`kube.WatchEvent` instance is yielded.  You can
        currently not control from "when" resources are being watched,
        other then from "now".  So be aware of any race conditions
        with watching.

        :retruns: An iterator of :class:`kube.WatchEvent` instances
           where the :attr:`kube.WatchEvent.item` attribute will be
           a :class:`kube.NodeItem`.

        :raises kube.APIError: For errors from the k8s API server.
        """
        nodelist = self.cluster.proxy.get('nodes')
        version = nodelist['metadata']['resourceVersion']
        jsonwatcher = self.cluster.proxy.watch('nodes', version=version)
        return _watch.ResourceWatcher(self.cluster, jsonwatcher, NodeItem)


class AddressType(enum.Enum):
    """Enumeration of the address types."""
    ExternalIP = 'ExternalIP'
    InternalIP = 'InternalIP'
    Hostname = 'Hostname'


class NodeItem(_base.ItemABC):
    """A node in the Kubernetes cluster.

    See http://kubernetes.io/docs/admin/node/ for details.

    :param cluster: The cluster this node belongs to.
    :type cluster: kube.Cluster
    :param raw: The raw data of the resource item.
    :type raw: pyrsistent.PMap

    :ivar cluster: The :class:`kube.Cluster` this node belongs to.

    :ivar raw: The raw decoded JSON object as returned by the API
       server.  This is a :class:`pyrsistent.PMap` instance.
    """
    _Address = collections.namedtuple('Address', ['type', 'addr'])

    def __init__(self, cluster, raw):  # pylint: disable=super-init-not-called
        self.cluster = cluster
        self.raw = raw

    @property
    def kind(self):
        """The Kubernetes resource kind of this resource.

        This is a :class:`kube.Kind` enum.
        """
        return _base.Kind.Node

    @property
    def meta(self):
        """The node's metadata as a :class:`kube.meta.ObjectMeta` instance.

        This provides access to the node's name, labels etc.
        """
        return _meta.ObjectMeta(self)

    def spec(self):
        """The spec of this node's resource item.

        This is the raw, decoded JSON data representing the spec of
        this node.
        """
        return pyrsistent.thaw(self.raw['spec'])

    @property
    def addresses(self):
        """An iterator of the addresses for this node.

        Each address is a namedtuple with ``(type, addr)`` as fields.
        Known types are in the :class:`kube.AddressType` enumeration.

        An empty list is returned if there are not yet any addresses
        associated with the node.
        """
        status = self.raw.get('status', {})
        for raw in status.get('addresses', []):
            type_ = AddressType(raw['type'])
            addr = ipaddress.ip_address(raw['address'])
            yield self._Address(type_, addr)

    @property
    def capacity(self):
        """The capacity of the node.

        CPU is expressed in cores and can use fractions of cores,
        while memory is expressed in bytes.
        """
        # See http://kubernetes.io/v1.1/docs/design/resources.html for
        # details on resources usage.  This needs to deal with custom
        # resources as well.  The current stub implementation does not
        # match well.
        raise NotImplementedError

    @property
    def conditions(self):
        """List of conditions."""
        raise NotImplementedError

    def watch(self):
        """Watch the node for changes.

        Only changes after the current version will be part of the
        iterator.  However it can not be guaranteed that *every*
        change is returned, if the current version is rather old some
        changes might no longer be available.

        :returns: An iterator of :class:`kube.WatchEvents` instances
           where the :attr:`kube.WatchEvent.item` attribute will be
           a :class:`kube.NodeItem` instance.
        :raises kube.APIError: For errors from the k8s API server.
        """
        return _util.watch_item(self)



"""Common exceptions for the kube package."""


class KubeError(Exception):
    """Base exception for all other exceptions from kube."""


class APIError(KubeError):
    """Error from the Kubernetes API server."""

    def __init__(self, response, message=None):
        self.response = response
        self.message = message
        self.status_code = response.status_code
        super().__init__(response, message)

    def __str__(self):
        try:
            api_message = self.response.json()['message']
        except Exception:       # pylint: disable=broad-except
            api_message = None
        if self.message and api_message:
            msg = self.message + '; ' + api_message
        else:
            msg = self.message or api_message or ''
        return msg + ' [HTTP {}]'.format(self.response.status_code)


class StatusError(KubeError):
    """The requested status item is unavailable."""


class NamespaceError(KubeError):
    """Invalid namespace.

    The resource does not support namespaces but was given one or a
    resource which requires a namespace was not given one.
    """



"""Some base classes and ABCs for kube.

While all implementations are supposed to at least register their
classes, subclassing is not really a aim.  With the exception of
:meth:`__repr__` most properties and methods are marked as abstract
and do not have a common implementation even if they could.  To share
code instead helper functions in ``kube._util`` should be used.
"""

import abc
import enum


class ViewABC(metaclass=abc.ABCMeta):
    """Represents a view to a collection of resources.

    All top-level resources in Kubernetes have a collection, resources
    of a ``*List`` kind, with some common functionality.  This ABC
    defines views to provide access to resources in collections in a
    uniform way.  Note that a view is not the same as the collection
    resource, e.g. collections resources have some metadata associated
    with them and exist at a particular point in time, they have a
    metadata.resourceVersion, which views do not have.

    It is always possible to create an instance of this without
    needing to do any requests to the real Kubernetes cluster.

    :param cluster: The cluster this resource list is part of.
    :type cluster: kube.Cluster
    :param namespace: The optional namespace this resource list is
       part of.  If the resource list is not part of a namespace this
       will be ``None`` which means it will be a view to all resources
       of a certain type, regardelss of their namespace.
    :type namespace: str

    :raises kube.NamespaceError: When a namespace is provided but the
       resource does not support one.

    :ivar kube.Cluster cluster: The cluster this resource list is part
       of.
    :ivar str namespace: The optional namespace this resource list is
       part of.  If the resource list is not part of a namespace this
       will be ``None``, including for resource lists which do not
       support namespaces.
    """

    @abc.abstractmethod
    def __init__(self, cluster, namespace=None):
        self.cluster = cluster
        self.namespace = namespace

    def __repr__(self):
        if self.namespace is None:
            return '<{0.__class__.__name__}>'.format(self)
        else:
            return ('<{0.__class__.__name__} namespace={0.namespace}>'
                    .format(self))

    @abc.abstractmethod         # pragma: nocover
    def __iter__(self):
        """Iterator of all resources in this collection.

        :raises kube.APIError: for errors from the k8s API server.
        """

    @abc.abstractproperty         # pragma: nocover
    def kind(self):
        """The kind of the underlying Kubernetes resource.

        This is a :class:`kube.Kind` enum.
        """

    @abc.abstractmethod         # pragma: nocover
    def fetch(self, name, namespace=None):
        """Retrieve the current version of a single resource item by name.

        If the view itself is associated with a namespace,
        ``self.namespace is not None``, then this will default to
        fetching the resource item from this namespace.  If the view
        is not associated with a namespace, ``self.namespace is
        None``, and the resource requires a namespace then a
        :class:`kube.NamespaceError` is raised.  Note that the
        ``default`` namespace is not automatically used in this case.

        :param str name: The name of the resource.
        :param str namespace: The namespace to fetch the resource
           from.

        :returns: A single instance representing the resource.

        :raises LookupError: If the resource does not exist.
        :raises kube.NamespaceError: For an invalid namespace, either
           because the namespace is required for this resource but not
           provided or the resource does not support namespaces and
           one was provided.
        :raises kube.APIError: For errors from the k8s API server.
        """

    @abc.abstractmethod         # pragma: nocover
    def filter(self, *, labels=None, fields=None):
        """Return an iterable of a subset of the resources.

        :param labels: A label selector expression.  This can either
           be a dictionary with labels which must match exactly, or a
           string label expression as understood by k8s itself.
        :type labels: dict or str
        :param fields: A field selector expression.  This can either
           be a dictionary with fields which must match exactly, or a
           string field selector as understood by k8s itself.
        :type fields: dict or str

        :returns: An iterator of :class:`kube.ItemABC` instances of
           the correct type for the resrouce which match the given
           selector.

        :raises ValueError: If an empty selector is used.  An empty
           selector is almost certainly not what you want.  Kubernetes
           treats an **empty** selector as *all* items and treats a
           **null** selector as *no* items.
        :raises kube.APIError: For errors from the k8s API server.
        """

    @abc.abstractmethod         # pragma: nocover
    def watch(self):
        """Watch for changes to any of the resources in the view.

        Whenever one of the resources in the view changes a new
        :class:`kube.WatchEvent` instance is yielded.  You can
        currently not control from "when" resources are being watched,
        other then from "now".  So be aware of any race conditions
        with watching.

        :returns: An iterator of :class:`kube.WatchEvent` instances.

        :raises kube.NamespaceError: Whe the namespace no longer
           exists.
        :raises kube.APIError: For errors from the k8s API server.
        """


class ItemABC(metaclass=abc.ABCMeta):
    """Representation for a kubernetes resource.

    All resources have some common attributes and API which must match
    this ABC.

    :param cluster: The cluster this resource is bound to.
    :type cluster: kube.Cluster
    :param raw: The decoded JSON representing the resource.
    :type raw: dict

    :ivar cluster: The :class:`kube.Cluster` instance this resource is
       bound to.
    :ivar raw: The raw decoded JSON representing the resource.  This
       behaves like a dict but is actually an immutable view of the
       dict.
    """

    @abc.abstractmethod
    def __init__(self, cluster, raw):
        self.cluster = cluster
        self.raw = raw

    def fetch(self):
        """Fetch the current version of the resource item.

        This will return a new instance of the current resource item
        at it's latest version.  This is useful to see any changes
        made to the object since it was last retrieved.

        :returns: An instance of the relevant :class:`ItemABC`
           subclass.

        :raises kube.APIError: For errors from the k8s API server.
        """
        data = self.cluster.proxy.get(self.meta.link)
        return self.__class__(self.cluster, data)

    def __repr__(self):
        if self.meta.namespace is not None:
            return ('<{0.__class__.__name__} {0.meta.name} '  # pylint: disable=missing-format-attribute
                    'namespace={0.meta.namespace}>' .format(self))
        else:
            return '<{0.__class__.__name__} {0.meta.name}>'.format(self)  # pylint: disable=missing-format-attribute

    @abc.abstractproperty       # pragma: nocover
    def kind(self):
        """The Kubernetes resource kind of the resource.

        This is a :class:`kube.Kind` enum.
        """

    @abc.abstractproperty       # pragma: nocover
    def meta(self):
        """The resource's metadata as a :class:`kube.ObjectMeta` instance."""

    @abc.abstractmethod       # pragma: nocover
    def spec(self):
        """The spec of this node's resource.

        This returns a copy of the raw, decodeed JSON data
        representing the spec of this resource which can be used to
        re-create the resource.
        """

    @abc.abstractmethod       # pragma: nocover
    def watch(self):
        """Watch the resource for changes.

        Whenever the resource changes a new :class:`kube.WatchEvent`
        is returned.

        :returns: an iterator of :class:`kube.WatchEvent` instances
           for the resource.
        """


class Kind(enum.Enum):
    """The kinds of Kubernetes resources."""
    Node = 'Node'
    NodeList = 'NodeList'
    Namespace = 'Namespace'
    NamespaceList = 'NamespaceList'
    Pod = 'Pod'
    PodList = 'PodList'
    ReplicaSet = 'ReplicaSet'
    ReplicaSetList = 'ReplicaSetList'
    Service = 'Service'
    ServiceList = 'ServiceList'
    Secret = 'Secret'
    SecretList = 'SecretList'



"""Interface for Secret resources."""

import base64
import enum

from kube import _base
from kube import _error
from kube import _meta
from kube import _util
from kube import _watch


class SecretView(_base.ViewABC):
    """View of the Secret resource items in the cluster.

    :param cluster: The cluster instance.
    :type cluster: kube.Cluster
    :param namespace: Limit the view to resource items in this
       namespace.
    :type namespace: str

    :ivar kube.Cluster cluster: The cluster instance this view is part of.
    :ivar str namespace: The namespace this view is contrained to, if
       the view is not bound to a namespace this will be ``None``.
    """

    def __init__(self, cluster, namespace=None):  # pylint: disable=super-init-not-called
        self.cluster = cluster
        self.namespace = namespace

    def __iter__(self):
        """Iterate over all Secret resource items in the view.

        :returns: An object for each secret in the view.
        :rtype: kube.SecretItem

        :raises kube.APIError: For errors from the k8s API server.
        """
        if self.namespace:
            data = self.cluster.proxy.get('namespaces',
                                          self.namespace, 'secrets')
        else:
            data = self.cluster.proxy.get('secrets')
        for item in data['items']:
            yield SecretItem(self.cluster, item)

    @property
    def kind(self):
        """The Kubernetes resource kind of the resource.

        This is a :class:`kube.Kind` enum.
        """
        return _base.Kind.SecretList

    def fetch(self, name, namespace=None):
        """Retrieve an individual secret by name.

        This returns the current verison of the resource item.

        If the view itself is associated with a namespace, the
        :attr:`namespace` attribute is not ``None``, then this will
        default to fetching the secret resource item from this
        namespace.  Otherwise the ``namespace`` parameter must be
        supplied.

        :param name: The name of the secret to retrive.
        :type name: str
        :param namespace: Use a namespace different from
           :attr:`namespace`.
        :type namespace: str

        :returns: A :class:`kube.SecretItem` instance.

        :raises LookupError: If the secret does not exist.
        :raises kube.NamespaceError: When no namespace was supplied.
        :raises kube.APIError: For errors from the k8s API server.
        """
        ns = namespace or self.namespace
        if not ns:
            raise _error.NamespaceError('Missing namespace')
        data = _util.fetch_resource(
            self.cluster, 'secrets', name, namespace=ns)
        return SecretItem(self.cluster, data)

    def filter(self, *, labels=None, fields=None):
        """Get a filtered iterator of secrets.

        :param labels: A label selector expression.  This can either
           be a dictionary with labels which must match exactly, or a
           string label expression as understood by k8s itself.
        :type labels: dict or str
        :param fields: A field selector expression.  This can either
           be a dictionary with fields which must match exactly, or a
           string field selector as understood by k8s itself.
        :type fields: dict or str

        :returns: an iterator of :class:`SecretItem`s that match the
           given selector.

        :raises ValueError: If an empty selector is used.  An empty
           selector is almost certainly not what you want.  Kubernetes
           treats an **empty** selector as *all* items and treats a
           **null** selector as *no* items.
        :raises kube.APIError: for errors from the k8s API server.
        """
        data_iter = _util.filter_list(self.cluster, 'secrets',
                                      labels=labels, fields=fields,
                                      namespace=self.namespace)
        for item in data_iter:
            yield SecretItem(self.cluster, item)

    def watch(self):
        """Watch the SecretList of this view for changes to items.

        Whenever one of the resource items in the view changes a new
        :class:`kube.WatchEvent` instance is yielded.  You can
        currently not control from "when" resources are being watched,
        other then from "now".  So be aware of any race conditions
        with watching.

        :returns: An iterator of :class:`kube.WatchEvent` instances
            where the :attr:`kube.WatchEvent.object` attribute will be
            a :class:`SecretItem` instance.

        :raises kube.NamespaceError: When the namespace no longer
           exists.
        :raises kube.APIError: For errors from the k8s API server.
        """
        if self.namespace:
            path = ['namespaces', self.namespace, 'secrets']
        else:
            path = ['secrets']
        svclist = self.cluster.proxy.get(*path)
        version = svclist['metadata']['resourceVersion']
        jsonwatcher = self.cluster.proxy.watch(*path, version=version)
        return _watch.ResourceWatcher(self.cluster, jsonwatcher, SecretItem)


class SecretType(enum.Enum):
    """Enumeration of secret types."""
    Opaque = 'Opaque'


class SecretItem(_base.ItemABC):
    """A Secret in the Kubernetes cluster.

    :param cluster: The cluster this Service exists in.
    :type cluster: kube.Cluster
    :param raw: The raw data of the resource item.
    :type raw: pyrsistent.PMap

    :ivar cluster: The :class:`kube.Cluster` this Secret exists in.
    :ivar raw: The raw decoded JSON object as returned by the API
       server.  This is a :class:`pyrsistent.PMap` instance.
    """
    SecretType = SecretType

    def __init__(self, cluster, raw):  # pylint: disable=super-init-not-called
        self.cluster = cluster
        self.raw = raw

    @property
    def kind(self):
        """The Kubernetes resource kind of this resource.

        This is a :class:`kube.Kind` enum.
        """
        return _base.Kind.Secret

    @property
    def meta(self):
        """The Secret's metadata as a :class:`kube.ObjectMeta` instance."""
        return _meta.ObjectMeta(self)

    def spec(self):
        """An empty dictionary.

        This is supposed to be the secret resource item's spec.  But
        secrets do not have a spec, so to still follow the
        :class:`kube.ItemABC` we return an empty dict.
        """
        return {}

    def watch(self):
        """Watch the secret for changes.

        Only changes after the current version will be part of the
        iterator.  However it can not be guaranteed that *every*
        change is returned, if the current version is rather old some
        changes might no longer be available.

        :returns: An iterator of :class:`kube.WatchEvent` instances
           where the :attr:`kube.WatchEvent.item` attribute will be
           a :class:`kube.SecretItem` instance.

        :raises kube.APIError: For errors from the k8s API server.
        """
        return _util.watch_item(self)

    @property
    def type(self):
        """The type of secret.

        There currently is only the "Opaque" type.
        """
        return self.SecretType(self.raw['type'])

    @property
    def data(self):
        """A mapping of the secret data.

        A copy of the secret data as a dict.  The keys are the names
        of the secrets as a (unicode) string, while the values are the
        secrets as bytestrings.

        Secret values are stored in a base64 encoding on the k8s
        master, but this is an implementation detail that this
        property takes care off for you.
        """
        data = {}
        for name, value in self.raw.get('data', {}).items():
            data[name] = base64.b64decode(value)
        return data



"""Interface for ReplicaSet resources.

These used to be called ReplicationController before.
"""

import pyrsistent

from kube import _base
from kube import _error
from kube import _meta
from kube import _util
from kube import _watch


class ReplicaSetView(_base.ViewABC):
    """View of the ReplicaSet resource items in the cluster.

    :param cluster: The cluster instance.
    :type cluster: kube.Cluster
    :param namespace: Limit the view to resource items in this
       namespace.
    :type namespace: str

    :ivar kube.Cluster cluster: The cluster instance this view is part of.
    :ivar str namespace: The namespace this view is contrained to, if
       the view is not bound to a namespace this will be ``None``.
    """

    def __init__(self, cluster, namespace=None):  # pylint: disable=super-init-not-called
        self.cluster = cluster
        self.namespace = namespace

    def __iter__(self):
        """Iterate over all Pod resource items in the view.

        :returns: An object for each pod in the view.
        :rtype: kube.PodItem

        :raises kube.APIError: For errors from the k8s API server.
        """
        if self.namespace:
            data = self.cluster.proxy.get('namespaces', self.namespace,
                                          'replicationcontrollers')
        else:
            data = self.cluster.proxy.get('replicationcontrollers')
        for item in data['items']:
            yield ReplicaSetItem(self.cluster, item)

    @property
    def kind(self):
        """The Kubernetes resource kind of the resource.

        This is a :class:`kube.Kind` enum.
        """
        return _base.Kind.ReplicaSetList

    def fetch(self, name, namespace=None):
        """Retrieve an individual ReplicaSet by name.

        This returns the current verison of the resource item.

        If the view itself is associated with a namespace, the
        :attr:`namespace` attribute is not ``None``, then this will
        default to fetching the ReplicaSet resource item from this
        namespace.  Otherwise the ``namespace`` parameter must be
        supplied.

        :param name: The name of the ReplicaSet to retrive.
        :type name: str
        :param namespace: Use a namespace different from
           :attr:`namespace`.
        :type namespace: str

        :returns: A :class:`kube.ReplicaSetItem` instance.

        :raises LookupError: If the ReplicaSet does not exist.
        :raises NamespaceError: When no namespace was supplied.
        :raises kube.APIError: For errors from the k8s API server.
        """
        ns = namespace or self.namespace
        if not ns:
            raise _error.NamespaceError('Missing namespace')
        data = _util.fetch_resource(
            self.cluster, 'replicationcontrollers', name, namespace=ns)
        return ReplicaSetItem(self.cluster, data)

    def filter(self, *, labels=None, fields=None):
        """Get a filtered iterator of ReplicaSets.

        :param labels: A label selector expression.  This can either
           be a dictionary with labels which must match exactly, or a
           string label expression as understood by k8s itself.
        :type labels: dict or str
        :param fields: A field selector expression.  This can either
           be a dictionary with fields which must match exactly, or a
           string field selector as understood by k8s itself.
        :type fields: dict or str

        :returns: an iterator of :class:`ReplicaSet`s that match the
           given selector.

        :raises ValueError: If an empty selector is used.  An empty
           selector is almost certainly not what you want.  Kubernetes
           treats an **empty** selector as *all* items and treats a
           **null** selector as *no* items.
        :raises kube.APIError: for errors from the k8s API server.
        """
        data_iter = _util.filter_list(self.cluster, 'replicationcontrollers',
                                      labels=labels, fields=fields,
                                      namespace=self.namespace)
        for item in data_iter:
            yield ReplicaSetItem(self.cluster, item)

    def watch(self):
        """Watch the ReplicaSetList of this view for changes to items.

        Whenever one of the resource items in the view changes a new
        :class:`kube.WatchEvent` instance is yielded.  You can
        currently not control from "when" resources are being watched,
        other then from "now".  So be aware of any race conditions
        with watching.

        :returns: An iterator of :class:`kube.WatchEvent` instances
            where the :attr:`kube.WatchEvent.object` attribute will be
            a :class:`ReplicaSetItem` instance.

        :raises kube.NamespaceError: When the namespace no longer
           exists.
        :raises kube.APIError: For errors from the k8s API server.
        """
        if self.namespace:
            path = ['namespaces', self.namespace, 'replicationcontrollers']
        else:
            path = ['replicationcontrollers']
        rslist = self.cluster.proxy.get(*path)
        version = rslist['metadata']['resourceVersion']
        jsonwatcher = self.cluster.proxy.watch(*path, version=version)
        return _watch.ResourceWatcher(self.cluster,
                                      jsonwatcher, ReplicaSetItem)


class ReplicaSetItem(_base.ItemABC):
    """A ReplicaSet in the Kubernetes cluster.

    A ReplicaSet, formerly known as a ReplicationController, is
    responsible for keeping a desired number of pods running.

    :param cluster: The cluster this ReplicaSet exists in.
    :type cluster: kube.Cluster
    :param raw: The raw data of the resource item.
    :type raw: pyrsistent.PMap

    :ivar cluster: The :class:`kube.Cluster` this ReplicaSet exists in.
    :ivar raw: The raw decoded JSON object as returned by the API
       server.  This is a :class:`pyrsistent.PMap` instance.
    """

    def __init__(self, cluster, raw):  # pylint: disable=super-init-not-called
        self.cluster = cluster
        self.raw = raw

    @property
    def kind(self):
        """The Kubernetes resource kind of this resource.

        This is a :class:`kube.Kind` enum.
        """
        return _base.Kind.ReplicaSet

    @property
    def meta(self):
        """The ReplicaSet's metadata as a :class:`kube.ObjectMeta` instance."""
        return _meta.ObjectMeta(self)

    def spec(self):
        """The spec of this ReplicaSet's resource item.

        This is a copy of the raw, decoded JSON data representing the
        spec of this ReplicaSet.
        """
        return pyrsistent.thaw(self.raw['spec'])

    def watch(self):
        """Watch the ReplicaSet for changes.

        Only changes after the current version will be part of the
        iterator.  However it can not be guaranteed that *every*
        change is returned, if the current version is rather old some
        changes might no longer be available.

        :returns: An iterator of :class:`kube.WatchEvent` instances
           where the :attr:`kube.WatchEvent.item` attribute will be
           a :class:`kube.ReplicaSetItem` instance.

        :raises kube.APIError: For errors from the k8s API server.
        """
        return _util.watch_item(self)

    @_util.statusproperty
    def observed_replicas(self):
        """The current number of replicas observed."""
        return self.raw['status']['replicas']

    @_util.statusproperty
    def observed_generation(self):
        """The (integer) generation of the ReplicaSet."""
        return self.raw['status']['observedGeneration']

    @_util.statusproperty
    def fully_labeled_replicas(self):
        """Number of pods which have an exact matching set of labels.

        This counts the pods which have the exact same set of labels
        as the labelselector of this replicaset.
        """
        return self.raw['status']['fullyLabeledReplicas']



"""Interface for namespaces."""

import enum

import pyrsistent

from kube import _base
from kube import _error
from kube import _meta
from kube import _util
from kube import _watch


class NamespaceView(_base.ViewABC):
    """View of all the Namespace resource items in the cluster.

    :param cluster: The cluster instance.
    :type cluster: kube.Cluster
    :param namespace: Limit the view to resource items in this
       namespace.  This is here for the :class:`kube.ViewABC`
       compatibility, namespaces can not be used for the NamespaceList
       resource.  A :class:`kube.NamespaceError` is raised when this
       is not ``None``.
    :type namespace: NoneType

    :raises kube.NamespaceError: If instantiated using a namespace.

    :ivar cluster: the :class:`kube.Cluster` instance.
    """

    def __init__(self, cluster, namespace=None):  # pylint: disable=super-init-not-called
        if namespace is not None:
            raise _error.NamespaceError(
                'NamespaceList does not support namespaces')
        self.cluster = cluster
        self.namespace = None

    def __iter__(self):
        """Iterate over all Namespace resource items in the cluster.

        :raises kube.APIError: For errors from the k8s API server.
        """
        data = self.cluster.proxy.get('namespaces')
        for item in data['items']:
            yield NamespaceItem(self.cluster, item)

    @property
    def kind(self):
        """The kind of the underlying Kubernetes list resource.

        This is a :class:`kube.Kind` enum.
        """
        return _base.Kind.NamespaceList

    def fetch(self, name, namespace=None):
        """Retrieve an individual Namespace resource item by name.

        This returns the current verison of the resource item.

        :param str name: The name of the namespace resource item to
           retrieve.
        :param str namespace: Must be *None* or a
           :class:`kube.NamespaceError` is raised.  Here only for
           compatibility with the ABC.

        :returns: A :class:`kube.NamespaceItem` instance.
        :raises LookupError: If the namespace does not exist.
        :raises kube.APIError: For errors from the k8s API server.
        :raises kube.NamespaceError: If a namespace is used.
        """
        if namespace is not None:
            raise _error.NamespaceError(
                'NodeList does not support namespaces')
        data = _util.fetch_resource(self.cluster, 'namespaces', name)
        return NamespaceItem(self.cluster, data)

    def filter(self, *, labels=None, fields=None):
        """Get a filtered iterator of namespaces.

        :param labels: A label selector expression.  This can either
           be a dictionary with labels which must match exactly, or a
           string label expression as understood by k8s itself.
        :type labels: dict or str
        :param fields: A field selector expression.  This can either
           be a dictionary with fields which must match exactly, or a
           string field selector as understood by k8s itself.
        :type fields: dict or str

        :returns: An iterator of :class:`NamespaceItem` instances
            which match the given selector.
        :raises ValueError: if an empty selector is used.  An empty
           selector is almost certainly not what you want.  Kubernetes
           treats an **empty** selector as *all* items and treats a
           **null** selector as *no* items.
        :raises kube.APIError: For errors from the k8s API server.
        """
        data_iter = _util.filter_list(self.cluster, 'namespaces',
                                      labels=labels, fields=fields)
        for data in data_iter:
            yield NamespaceItem(self.cluster, data)

    def watch(self):
        """Watch the NamespaceList of this view for changes to items.

        Whenever one of the Namespace resource items in the view
        changes a new :class:`kube.WatchEvent` instance is yielded.
        You can currently not control from "when" resources are being
        watched, other then from "now".  So be aware of any race
        conditions with watching.

        :returns: An iterator of :class:`kube.WatchEvent` instances
           where the :attr:`kube.WatchEvent.item` attribute will be a
           :class:`kube.NamespaceItem`.
        :raise kube.APIError: For errors from the k8s API server.
        """
        nslist = self.cluster.proxy.get('namespaces')
        version = nslist['metadata']['resourceVersion']
        jsonwatcher = self.cluster.proxy.watch('namespaces', version=version)
        return _watch.ResourceWatcher(self.cluster, jsonwatcher, NamespaceItem)


class NamespacePhase(enum.Enum):
    """Enumeration of all possible namespace phases.

    This is aliased to :attr:`NamespaceResource.NamespacePhase`
    for convenience.
    """
    Active = 'Active'
    Terminating = 'Terminating'


class NamespaceItem(_base.ItemABC):
    """A namespace in the Kubernetes cluster.

    See http://kubernetes.io/docs/admin/namespaces/ for details.

    :cvar NamespacePhase: Convenience alias of :class:`NamespacePhase`.

    :ivar cluster: the :class:`kube.Cluster` instance.
    :ivar raw: the raw JSON-decoded representation of the namespace.
    """

    NamespacePhase = NamespacePhase

    def __init__(self, cluster, raw):  # pylint: disable=super-init-not-called
        self.cluster = cluster
        self.raw = raw

    @property
    def kind(self):
        """The kind of the underlying Kubernetes resource.

        This is a :class:`kube.Kind` enum.
        """
        return _base.Kind.Namespace

    @property
    def meta(self):
        """Namespace's metadata as a :class:`kube.ObjectMeta`."""
        return _meta.ObjectMeta(self)

    def spec(self):
        """A copy of the spec of this namespace's resource.

        This is the raw, decoded JSON data representing the spec of
        this namespace.
        """
        return pyrsistent.thaw(self.raw['spec'])

    @property
    def phase(self):
        """Phase of the namespace as a :class:`kube.NamespacePhase`."""
        return NamespacePhase(self.raw['status']['phase'])

    def watch(self):
        """Watch the namespace for changes.

        Only changes after the current version will be part of the
        iterator.  However it can not be guaranteed that *every*
        change is returned, if the current version is rather old some
        changes might no longer be available.

        :retruns: An iterator of :class:`kube.WatchEvents` instances
           where the :attr:`kube.WatchEvent.item` attribute will be
           a :class:`kube.NamespaceItem` instance..
        :raises kube.APIError: For errors from the k8s API server.
        """
        return _util.watch_item(self)



"""Interface for pod resources."""

import enum
import ipaddress

import pyrsistent

from kube import _base
from kube import _error
from kube import _meta
from kube import _util
from kube import _watch


class PodView(_base.ViewABC):
    """View of the Pod resource items in the cluster.

    :param cluster: The cluster instance.
    :type cluster: kube.Cluster
    :param namespace: Limit the view to resource items in this
       namespace.
    :type namespace: str

    :ivar kube.Cluster cluster: The cluster instance this view is part of.
    :ivar str namespace: The namespace this view is contrained to, if
       the view is not bound to a namespace this will be ``None``.
    """

    def __init__(self, cluster, namespace=None):  # pylint: disable=super-init-not-called
        self.cluster = cluster
        self.namespace = namespace

    def __iter__(self):
        """Iterate over all Pod resource items in the view.

        :returns: An object for each pod in the view.
        :rtype: kube.PodItem

        :raises kube.APIError: For errors from the k8s API server.
        """
        if self.namespace:
            data = self.cluster.proxy.get('namespaces', self.namespace, 'pods')
        else:
            data = self.cluster.proxy.get('pods')
        for item in data['items']:
            yield PodItem(self.cluster, item)

    @property
    def kind(self):
        """The Kubernetes resource kind of the resource.

        This is a :class:`kube.Kind` enum.
        """
        return _base.Kind.PodList

    def fetch(self, name, namespace=None):
        """Retrieve an individual pod by name.

        This returns the current verison of the resource item.

        If the view itself is associated with a namespace, the
        :attr:`namespace` attribute is not ``None``, then this will
        default to fetching the pod resource item from this namespace.
        Otherwise the ``namespace`` parameter must be supplied.

        :param name: The name of the pod to retrive.
        :type name: str
        :param namespace: Use a namespace different from
           :attr:`namespace`.
        :type namespace: str

        :returns: A :class:`kube.PodItem` instance.

        :raises LookupError: If the pod does not exist.
        :raises NamespaceError: When no namespace was supplied.
        :raises kube.APIError: For errors from the k8s API server.
        """
        ns = namespace or self.namespace
        if not ns:
            raise _error.NamespaceError('Missing namespace')
        data = _util.fetch_resource(self.cluster, 'pods', name, namespace=ns)
        return PodItem(self.cluster, data)

    def filter(self, *, labels=None, fields=None):
        """Get a filtered iterator of pods.

        :param labels: A label selector expression.  This can either
           be a dictionary with labels which must match exactly, or a
           string label expression as understood by k8s itself.
        :type labels: dict or str
        :param fields: A field selector expression.  This can either
           be a dictionary with fields which must match exactly, or a
           string field selector as understood by k8s itself.
        :type fields: dict or str

        :returns: an iterator of :class:`Pod`s that match the given selector.
        :raises ValueError: If an empty selector is used.  An empty
           selector is almost certainly not what you want.  Kubernetes
           treats an **empty** selector as *all* items and treats a
           **null** selector as *no* items.
        :raises kube.APIError: for errors from the k8s API server.
        """
        data_iter = _util.filter_list(self.cluster, 'pods',
                                      labels=labels, fields=fields,
                                      namespace=self.namespace)
        for item in data_iter:
            yield PodItem(self.cluster, item)

    def watch(self):
        """Watch the PodList of this view for changes to items.

        Whenever one of the resource items in the view changes a new
        :class:`kube.WatchEvent` instance is yielded.  You can
        currently not control from "when" resources are being watched,
        other then from "now".  So be aware of any race conditions
        with watching.

        :returns: An iterator of :class:`kube.WatchEvent` instances
            where the :attr:`kube.WatchEvent.object` attribute will be
            a :class:`PodItem` instance.

        :raises kube.NamespaceError: When the namespace no longer
           exists.
        :raises kube.APIError: For errors from the k8s API server.
        """
        if self.namespace:
            path = ['namespaces', self.namespace, 'pods']
        else:
            path = ['pods']
        podlist = self.cluster.proxy.get(*path)
        version = podlist['metadata']['resourceVersion']
        jsonwatcher = self.cluster.proxy.watch(*path, version=version)
        return _watch.ResourceWatcher(self.cluster, jsonwatcher, PodItem)


class PodPhase(enum.Enum):
    """Enumeration of all possible pod phases.

    This is aliased to :attr:`Pod.PodPhase` for convenience.
    """
    Pending = 'Pending'
    Running = 'Running'
    Succeeded = 'Succeeded'
    Failed = 'Failed'
    Unknown = 'Unknown'


class PodItem(_base.ItemABC):
    """A pod in the Kubernetes cluster.

    Each pod contains a number of containers and volumes which are executed
    on a node within the cluster. A pod may exist in a namespace. Pods are
    typically managed by a controller such as a replication controller or
    job.

    :param cluster: The cluster this pod exists in.
    :type cluster: kube.Cluster
    :param raw: The raw data of the resource item.
    :type raw: pyrsistent.PMap

    :cvar PodPhase: Convenience alias of :class:`PodPhase`.

    :ivar cluster: The :class:`kube.Cluster` this pod exists in.
    :ivar raw: The raw decoded JSON object as returned by the API
       server.  This is a :class:`pyrsistent.PMap` instance.
    """
    PodPhase = PodPhase

    def __init__(self, cluster, raw):  # pylint: disable=super-init-not-called
        self.cluster = cluster
        self.raw = raw

    @property
    def kind(self):
        """The Kubernetes resource kind of this resource.

        This is a :class:`kube.Kind` enum.
        """
        return _base.Kind.Pod

    @property
    def meta(self):
        """Pod's metadata as a :class:`kube.ObjectMeta`."""
        return _meta.ObjectMeta(self)

    def spec(self):
        """The spec of this pod's resource item.

        This a copy of the raw, decoded JSON data representing the
        spec of this pod.
        """
        return pyrsistent.thaw(self.raw['spec'])

    @_util.statusproperty
    def phase(self):
        """Phase of the pod as a :class:`kube.PodPhase`.

        :raises kube.StatusError: If this status item is not present.
        """
        return PodPhase(self.raw['status']['phase'])

    @_util.statusproperty
    def start_time(self):
        """Start the pod was started as a :class:`datetime.datetime`.

        :raises kube.StatusError: If this status item is not present.
        """
        return _util.parse_rfc3339(self.raw['status']['startTime'])

    @_util.statusproperty
    def ip(self):               # pylint: disable=invalid-name
        """IP address of the pod within the cluster.

        This may be as a :class:`ipaddress.IPv4Address` or a
        :class:`ipaddress.IPV6Address`.

        :raises kube.StatusError: If this status item is not present.
        """
        return ipaddress.ip_address(self.raw['status']['podIP'])

    @_util.statusproperty
    def host_ip(self):
        """IP address of the pod's host within the cluster.

        This  may be as a :class:`ipaddress.IPv4Address` or a
        :class:`ipaddress.IPv6Address`.

        :raises kube.StatusError: If this status item is not present.
        """
        return ipaddress.ip_address(self.raw['status']['hostIP'])

    @_util.statusproperty
    def message(self):
        """Human readable message explaining the pod's state.

        :raises kube.StatusError: If this status item is not present.
        """
        return self.raw['status']['message']

    @_util.statusproperty
    def reason(self):
        """PascalCase string explaining the pod's state.

        :raises kube.StatusError: If this status item is not present.
        """
        return self.raw['status']['reason']

    @property
    def containers(self):
        """Iterate over all :class:`Container`s in the pod."""
        statuses = self.raw['status'].get('containerStatuses', [])
        for container_status in statuses:
            yield Container(self, container_status)

    def watch(self):
        """Watch the pod for changes.

        Only changes after the current version will be part of the
        iterator.  However it can not be guaranteed that *every*
        change is returned, if the current version is rather old some
        changes might no longer be available.

        :returns: An iterator of :class:`kube.WatchEvent` instances
           where the :attr:`kube.WatchEvent.item` attribute will be
           a :class:`kube.PodItem` instance.

        :raises kube.APIError: For errors from the k8s API server.
        """
        return _util.watch_item(self)


class Container:
    """A container inside a pod.

    Containers live inside a pod and may be restarted inside this pod
    as controlled by the restart policy set on the pod.

    :param pod: The pod the container is part off.
    :type pod: PodItem
    :param raw: The JSON-decoded object describing the status of the
       container.
    :type raw: pyrsistent.PMap

    :ivar pod: The :class:`PodItem` instance the container is bound to.
    :ivar raw: The raw JSON-decoded object representing the container.
    """

    def __init__(self, pod, raw):
        self.pod = pod
        self.raw = raw

    def __repr__(self):
        return "<{0.__class__.__name__} {0.name!r} of {0.pod}>".format(self)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.raw == other.raw
        else:
            return NotImplemented

    @property
    def name(self):
        """The name of the container as a string."""
        return self.raw['name']

    @_util.statusproperty
    def state(self):
        """Current state of the container.

        This is represented by a :class:`ContainerState` instance.

        :raises kube.StatusError: If this status item is not present.
        """
        state = ContainerState(self.raw['state'])
        return state

    @_util.statusproperty
    def last_state(self):
        """Previous state of the container, if known.

        This is represented by a :class:`ContainerState` instance.

        :raises kube.StatusError: If this status item is not present.
        """
        if 'lastState' in self.raw and self.raw['lastState']:
            return ContainerState(self.raw['lastState'])
        else:
            raise _error.StatusError('No previous container state available')

    @_util.statusproperty
    def ready(self):
        """Boolean indicating if the container passed it's readyness probe.

        :raises kube.StatusError: If this status item is not present.
        """
        return bool(self.raw['ready'])

    @_util.statusproperty
    def restart_count(self):
        """The number of times the container was restarted as an integer.

        Note that this is currently not always accurate, it counts the
        number of dead containers which have not yet been removed.
        This means the gargage collection of containers caps this
        number at 5.
        """
        return int(self.raw['restartCount'])

    @_util.statusproperty
    def image(self):
        """The image the container is running.

        For Docker this is normally the repository name with tag
        appended.
        """
        return self.raw['image']

    @_util.statusproperty
    def image_id(self):
        """The ImageID of the container's image.

        For Docker this is in the ``docker://<hex_id>`` format.
        """
        return self.raw['imageID']

    @property
    def id(self):               # pylint: disable=invalid-name
        """The ID of the running container.

        For Docker this is in the ``docker://<hex_id>`` format.
        """
        return self.raw['containerID']


class ContainerState:
    """The state of a container within a pod.

    A container can be in one of three states: *running*, *waiting* or
    *terminated*.  This class provides a uniform interface to all
    states and their associated details.  Not all fields are always
    valid for each state so they can all raise an
    :exc:`kube.StatusError` when they are not available or not
    applicable.

    The overall state of the container is available both as a string
    in the :attr:`state` attribute as well as booleans in the
    :attr:`waiting`, :attr:`running` and :attr:`terminated`
    attributes.

    :param raw: The raw JSON-decoded ``v1.ContainerState`` API object
        as exposed by ``v1.ContainerStatus`` objects.
    :type raw: pyrsistent.PMap

    :ivar raw: The raw JSON-decoded object representing the container state.
    """

    def __init__(self, raw):
        # Sometimes k8s gets it wrong and ends up with multiple states
        # instead of just the one.  In this case the state in
        # lastState will still be present in the existing state.  So
        # we prefer states in a certain order.
        states = set(raw.keys())
        if len(states) > 1:
            if states == {'waiting', 'terminated'}:
                self._state = 'waiting'
            elif states == {'waiting', 'running'}:
                self._state = 'running'
            elif states == {'running', 'terminated'}:
                self._state = 'terminated'
            else:
                raise ValueError('Unknown state combination')
        else:
            self._state = states.pop()
        self._data = raw[self._state]
        self.raw = raw

    def __repr__(self):
        return '<{0.__class__.__name__} {0._state}>'.format(self)

    @property
    def waiting(self):
        """Boolean indicating if the container is waiting."""
        return self._state == 'waiting'

    @property
    def running(self):
        """Boolean indicating if the container is running."""
        return self._state == 'running'

    @property
    def terminated(self):
        """Boolean indicating if the container has been terminated."""
        return self._state == 'terminated'

    @_util.statusproperty
    def reason(self):
        """Brief reason explaining the container's state (str).

        This is normally a CamelCased message ID.

        Available for *waiting* and *terminated* states.

        :raises kube.StatusError: When this is not provided.
        """
        return self._data['reason']

    @_util.statusproperty
    def message(self):
        """Message regarding the container's state (str).

        Available for *waiting* and *terminated* states.

        :raises kube.StatusError: When this is not provided.
        """
        return self._data['message']

    @_util.statusproperty
    def started_at(self):
        """The time the container was started or restarted (datetime.datetime).

        Available for the *running* state.

        :raises kube.StatusError: When this is not provided.
        """
        return _util.parse_rfc3339(self._data['startedAt'])

    @_util.statusproperty
    def exit_code(self):
        """Exit code of the container (int).

        Available for the *terminated* state.

        :raises kube.StatusError: When this is not provided.
        """
        return self._data['exitCode']

    @_util.statusproperty
    def signal(self):
        """Last signal sent to the container, if known (int).

        Not all terminated containers can be expected to have this.

        .. warning::
            The signal is identified numerically, however these signal
            numbers are not portable therefore it's ill-advised to attempt
            to compare this value with the constants provided by the
            built-in :mod:`singal` module.

        Available for the *terminated* state.

        :raises kube.StatusError: When this is not provided.
        """
        return self._data['signal']

    @_util.statusproperty
    def finished_at(self):
        """The time the container was terminated (datetime.datetime).

        Available for the *terminated* state.

        :raises kube.StatusError: When this is not provided.
        """
        return _util.parse_rfc3339(self._data['finishedAt'])

    @_util.statusproperty
    def container_id(self):
        """The container ID of the terminated container.
        Available for the *terminated* state.

        :raises kube.StatusError: When this is not provided.
        """
        return self._data['containerID']



"""Python setuptools build script."""

import sys

import setuptools


__version__ = '0.8.0'


INSTALL_REQUIRES = [
    'requests',
    'http-parser',
    'pyrsistent',
]
if sys.version_info < (3, 4):
    INSTALL_REQUIRES.append('enum34')


CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
    'Topic :: Software Development :: Libraries',
    'Programming Language :: Python :: 3 :: Only',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
]


setuptools.setup(
    name='kube',
    version=__version__,
    author='Floris Bruynooghe',
    author_email='flub@cobe.io',
    license='LGPLv3',
    url='http://bitbucket.org/cobeio/kube',
    description='Opinionated interface for the Kubernetes API',
    packages=setuptools.find_packages(),
    install_requires=INSTALL_REQUIRES,
    classifiers=CLASSIFIERS,
    keywords='kubernetes k8s watch',
)
